#!/usr/bin/env python
##############################################################################
# (c) Crown copyright 2019 Met Office. All rights reserved.
# The file LICENCE, distributed with this code, contains details of the terms
# under which the code may be used.
##############################################################################
'''
Classes relating to styles made up of rules.
'''

from abc import ABCMeta
import logging

import stylist.issue
import stylist.rule


class Style(object):
    # pylint: disable=too-few-public-methods
    '''
    Abstract parent of all style lists.
    '''
    __metaclass__ = ABCMeta

    def __init__(self, rules):
        if not isinstance(rules, list):
            rules = [rules]
        self._rules = rules

    def list_rules(self):
        '''
        Gets a list of the rules which make up this style.
        '''
        return [rule.__class__.__name__ for rule in self._rules]

    def check(self, source):
        '''
        Applies every rule in this style to the parse tree.

        source Program source code as an stylist.source.Source derived
               object.
        '''
        logging.getLogger(__name__).info('Style: ' + self.__class__.__name__)
        issues = []
        for rule in self._rules:
            issues.extend(rule.examine(source))
        return issues


class LFRicStyle(Style):
    # pylint: disable=too-few-public-methods
    '''
    LFRic project's list of rules.
    '''
    def __init__(self):
        rules = [stylist.rule.FortranCharacterset(),
                 stylist.rule.TrailingWhitespace(),
                 stylist.rule.MissingImplicit('none'),
                 stylist.rule.MissingOnly(ignore=['pfunit_mod', 'xios'])]
        super(LFRicStyle, self).__init__(rules)



#!/usr/bin/env python
##############################################################################
# (c) Crown copyright 2019 Met Office. All rights reserved.
# The file LICENCE, distributed with this code, contains details of the terms
# under which the code may be used.
##############################################################################
'''
Tool for checking code style.
'''

import argparse
import logging
import os.path
import sys

from stylist.engine import CheckEngine
from stylist.source import CPreProcessor, CSource, FortranPreProcessor, \
                           FortranSource, PFUnitProcessor, \
                           SourceFactory
from stylist.style import LFRicStyle


def parse_cli():
    '''
    Parse the command line. Returns a dictionary of arguments.
    '''
    description = 'Perform various code style checks on source code.'
    cli_parser = argparse.ArgumentParser(add_help=False,
                                         description=description)
    cli_parser.add_argument('-help', '-h', '--help', action='help')
    cli_parser.add_argument('-verbose', '-v', action="store_true",
                            help='Produce a running commentary on progress')
    message = 'Add a mapping between file extension and language'
    cli_parser.add_argument('-map-extension',
                            metavar='EXTENSION:LANGUAGE[:PREPROCESSOR]',
                            dest='map_extension',
                            default=[],
                            action='append',
                            help=message)
    cli_parser.add_argument('source', metavar='FILE', nargs='+',
                            help='Filename of source file or directory')
    return cli_parser.parse_args()


_LANGUAGE_MAP = {'c': CSource,
                 'cxx': CSource,
                 'fortran': FortranSource}
_PREPROCESSOR_MAP = {'cpp': CPreProcessor,
                     'fpp': FortranPreProcessor,
                     'pfp': PFUnitProcessor}


def _add_extensions(additional_extensions):
    # This application always expects pFUnit source to be present so it adds
    # a rule for that.
    #
    SourceFactory.add_extension('pf',
                                FortranSource,
                                PFUnitProcessor)
    SourceFactory.add_extension('PF',
                                FortranSource,
                                FortranPreProcessor,
                                PFUnitProcessor)
    # PSyclone algorithms are also expected. By design these are actually
    # syntactically correct Fortran so they are treated exactly as Fortran
    # files are. In particular, they are treated like .F90 files as
    # preprocessor directives may appear in them.
    #
    SourceFactory.add_extension('x90',
                                FortranSource,
                                FortranPreProcessor)

    for mapping in additional_extensions:
        extension, language, preprocessors = mapping.split(':', 2)
        preprocessors = preprocessors.split(':')
        preproc_objects = [_PREPROCESSOR_MAP[ident.lower()]
                           for ident in preprocessors]
        SourceFactory.add_extension(extension.lower(),
                                    _LANGUAGE_MAP[language.lower()],
                                    *preproc_objects)


def process(arguments):
    '''
    Examines files for style compliance.

    Any directories specified will be descended looking for files to examine.
    '''
    logger = logging.getLogger('stylist')
    logger.addHandler(logging.StreamHandler(sys.stdout))

    if arguments.verbose:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    _add_extensions(arguments.map_extension)

    styles = [LFRicStyle()]
    engine = CheckEngine(styles)

    hot_extensions = SourceFactory.get_extensions()

    candidates = arguments.source
    issues = []
    while candidates:
        filename = candidates.pop(0)
        if os.path.isdir(filename):
            for leaf in os.listdir(filename):
                leaf_filename = os.path.join(filename, leaf)
                _, extension = os.path.splitext(leaf_filename)
                if extension[1:] in hot_extensions \
                   or os.path.isdir(leaf_filename):
                    candidates.append(leaf_filename)
        else:  # Is a file
            issues.extend(engine.check(filename))

    for issue in issues:
        print(str(issue), file=sys.stderr)
    if (len(issues) > 0) or arguments.verbose:
        print('Found {number} issues'.format(number=len(issues)))

    if issues:
        sys.exit(1)


def main():
    '''Main entry point.'''
    return process(parse_cli())



"""Setup Stylist for distribution on PyPI."""

from ast import literal_eval
import os
from setuptools import setup, find_packages

# Get the long description from the README file
HERE = os.path.dirname(__file__)
with open(os.path.join(HERE, 'README.md'), encoding='utf-8',) as handle:
    LONG_DESCRIPTION = handle.read()
with open(
    os.path.join(HERE, 'source', 'stylist', '__init__.py'),
    encoding='utf-8',
) as handle:
    for line in handle:
        items = line.split('=', 1)
        if items[0].strip() == '__version__':
            VERSION = literal_eval(items[1].strip())
            break
    else:
        raise RuntimeError('Cannot determine package version.')

setup(
    name='stylist',
    version=VERSION,
    description=(
        'Extensible code style checker'
        ' currently supporting Fortran, PSyclone DSL, etc'
    ),
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/MetOffice/stylist',
    author='Met Office',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Quality Assurance',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='linter fortran psyclone',
    package_dir={'': 'source'},
    packages=find_packages(where='source'),
    python_requires='>=3.6, <4',
    install_requires=['fparser'],
    extras_require={
        'dev': ['check-manifest', 'flake8'],
        'test': ['pytest', 'pytest-cov'],
    },
    entry_points={
        'console_scripts': ['stylist=stylist.__main__:main'],
    },
    project_urls={
        'Bug Reports': 'https://github.com/MetOffice/stylist/issues',
        'Source': 'https://github.com/MetOffice/stylist/',
    },
)



#!/usr/bin/env python
##############################################################################
# (c) Crown copyright 2019 Met Office. All rights reserved.
# The file LICENCE, distributed with this code, contains details of the terms
# under which the code may be used.
##############################################################################
'''
Manages source code in various flavours.
'''

from abc import ABCMeta, abstractmethod
import re
import os.path

import fparser.common.readfortran as readfortran
import fparser.two.Fortran2003
from fparser.two.parser import ParserFactory
from fparser.two.utils import FparserException


class SourceText(object, metaclass=ABCMeta):
    # pylint: disable=too-few-public-methods
    '''
    Handles source code at the text level. Makes use of the decorator pattern
    to perform text level preprocessing.
    '''
    @abstractmethod
    def get_text(self):
        '''
        Gets the source file as a string.
        '''
        raise NotImplementedError()


class SourceFileReader(SourceText):
    # pylint: disable=too-few-public-methods
    '''
    Reads source from a file.
    '''
    def __init__(self, sourceFile):
        '''
        Constructor.
        Accepts either a filename or file-like object.
        '''
        if isinstance(sourceFile, str):
            with open(sourceFile, 'rt') as handle:
                self._cache = handle.read()
        else:
            self._cache = sourceFile.read()

    def get_text(self):
        return self._cache


class SourceStringReader(SourceText):
    # pylint: disable=too-few-public-methods
    '''
    Reads source from a string.
    '''
    def __init__(self, source_string):
        '''
        Constructor.
        '''
        self._source_string = source_string

    def get_text(self):
        return self._source_string


class TextProcessor(SourceText, metaclass=ABCMeta):
    # pylint: disable=too-few-public-methods, abstract-method
    '''
    Preprocessor decorators inherit from this.

    @param source The SourceText object which this object decorates.
    '''
    def __init__(self, source):
        self._source = source


class CPreProcessor(TextProcessor):
    # pylint: disable=too-few-public-methods
    '''
    Strips out preprocessor directives.

    It is assumed that you want to syntax check all the Source so all
    conditional directives such as '#ifdef' are stripped out.

    TODO: Currently all other directives are stripped out as well. This means
          that macros used to inject source are not likely to parse.
    '''
    _CONDITIONAL_DIRECTIVE_PATTERN = re.compile(r'^(\s*)(#if(def|\s+)*)$',
                                                re.MULTILINE)
    _OTHER_DIRECTIVE_PATTERN = re.compile(r'^(\s*)(#.*)$', re.MULTILINE)

    def get_text(self):
        '''
        Strips preprocessor directives from the text.
        '''
        text = self._source.get_text()
        text = self._CONDITIONAL_DIRECTIVE_PATTERN.sub(r'\1// \2', text)
        text = self._OTHER_DIRECTIVE_PATTERN.sub(r'\1// \2', text)
        return text


class FortranPreProcessor(TextProcessor):
    # pylint: disable=too-few-public-methods
    '''
    Strips out preprocessor directives.

    It is assumed that you want to syntax check all the Source so all
    conditional directives such as '#ifdef' are stripped out.

    TODO: Currently all other directives are stripped out as well. This means
          that macros used to inject source are not likely to parse.
    '''
    _CONDITIONAL_DIRECTIVE_PATTERN = re.compile(r'^(\s*)(#if(def|\s+)*)$',
                                                re.MULTILINE)
    _OTHER_DIRECTIVE_PATTERN = re.compile(r'^(\s*)(#.*)$', re.MULTILINE)

    def get_text(self):
        '''
        Strips preprocessor directives from the text.
        '''
        text = self._source.get_text()
        text = self._CONDITIONAL_DIRECTIVE_PATTERN.sub(r'\1! \2', text)
        text = self._OTHER_DIRECTIVE_PATTERN.sub(r'\1! \2', text)
        return text


class PFUnitProcessor(TextProcessor):
    # pylint: disable=too-few-public-methods
    '''
    Strips out pFUnit directives.
    '''
    _DIRECTIVE_PATTERN = re.compile(r'^(\s*)(@\w+.*)$', re.MULTILINE)

    def get_text(self):
        '''
        Strips processor directives from the text.
        '''
        text = self._source.get_text()
        text = self._DIRECTIVE_PATTERN.sub(r'\1! \2', text)
        return text


class SourceTree(object, metaclass=ABCMeta):
    '''
    Abstract parent of all actual language source files.
    '''
    def __init__(self, text):
        if not isinstance(text, SourceText):
            raise Exception('text argument must derive from SourceText.')

        self._text = text
        self._tree = None
        self._tree_error = 'Not parsed yet'

    @abstractmethod
    def get_tree(self):
        '''
        Gets a parse-tree representation of the source file.
        '''
        raise NotImplementedError()

    @abstractmethod
    def get_tree_error(self):
        '''
        Gets any errors raised while building the parse tree.
        '''
        raise NotImplementedError()

    def get_text(self):
        '''
        Gets the original source text.
        '''
        return self._text.get_text()


class FortranSource(SourceTree):
    '''
    Holds a Fortran source file as both a text block and parse tree.
    '''
    def get_tree(self):
        if not self._tree:
            # We don't use the tree directly. Instead we let all the decorators
            # have a go first.
            reader = readfortran.FortranStringReader(self._text.get_text(),
                                                     ignore_comments=False)
            fortran_parser = ParserFactory().create(std='f2008')
            try:
                self._tree = fortran_parser(reader)
                self._tree_error = None
            except FparserException as ex:
                self._tree = None
                self._tree_error = str(ex)
        return self._tree

    def get_tree_error(self):
        return self._tree_error

    def get_first_statement(self, root=None):
        '''
        Gets the first "statement" part of the syntax tree or part thereof.
        '''
        if not root:
            root = self._tree.content

        root = [root]

        while root:
            candidate = root.pop(0)
            if isinstance(candidate, fparser.two.Fortran2003.StmtBase):
                return candidate

            if isinstance(candidate, fparser.two.Fortran2003.BlockBase):
                root.extend(candidate.content)
            elif isinstance(candidate, fparser.two.Fortran2003.SequenceBase):
                root.extend(candidate.items)
            elif isinstance(candidate, fparser.two.Fortran2003.Comment):
                pass
            else:
                message = 'Unexpected candidate type: {0}'
                raise Exception(message.format(candidate.__class__.__name__))
        return None

    def path(self, path, root=None):
        # pylint: disable=too-many-branches
        '''
        Returns the tree node described by the given path. The path may be
        presented as either a list of node name strings or as a '/' delimited
        string of node names.

        If the path is not found in the tree, returns an empty list.
        '''
        if isinstance(path, str):
            path = path.split('/')
        else:
            path = list(path)

        found = []
        if root:
            next_candidates = root
        else:
            next_candidates = list(self.get_tree().content)

        while path:
            node_name = path.pop(0)
            # pylint: disable=eval-used
            node_class = eval('fparser.two.Fortran2003.' + node_name)

            candidates = next_candidates
            next_candidates = []
            while candidates:
                candidate = candidates.pop(0)
                candidate_name = candidate.__class__.__name__
                node_subclasses = node_class.subclass_names
                # Only allow comments to be considered if we are looking for
                # a comment.
                if not isinstance(node_class, fparser.two.Fortran2003.Comment):
                    if 'Comment' in node_subclasses:
                        node_subclasses.remove('Comment')

                if self._ast_match(candidate, node_class):
                    if not path:  # Bottom of the path
                        found.append(candidate)

                    if isinstance(candidate,
                                  fparser.two.Fortran2003.BlockBase):
                        next_candidates.extend(candidate.content)
                    elif isinstance(candidate,
                                    fparser.two.Fortran2003.SequenceBase):
                        next_candidates.extend(candidate.items)
                    elif isinstance(candidate,
                                    fparser.two.Fortran2003.StmtBase):
                        pass
                    elif isinstance(candidate,
                                    fparser.two.Fortran2003.Comment):
                        pass
                    else:
                        message = 'Unexpected candidate type: {0}'
                        raise Exception(message.format(candidate_name))
        return found

    def find_all(self, find_node, root=None):
        '''
        Returns a generator which loops over all instances if the specified
        parse element below the root.

        The search descends the tree but that descent is terminated by a match.
        '''
        if root:
            candidates = [root]
        else:
            candidates = list(self.get_tree().content)

        while candidates:
            candidate = candidates.pop(0)
            if self._ast_match(candidate, find_node):
                yield candidate
            else:  # If we found a thing we don't descend into it
                if isinstance(candidate, fparser.two.Fortran2003.BlockBase):
                    candidates.extend(candidate.content)
                elif isinstance(candidate,
                                fparser.two.Fortran2003.SequenceBase):
                    candidates.extend(candidate.items)
                else:
                    pass

    @staticmethod
    def _ast_match(candidate, saught_class):
        if candidate.__class__.__name__ == saught_class.__name__:
            return True

        considering = [saught_class]
        for consideration in considering:
            if candidate.__class__.__name__ == consideration.__name__:
                return True
            # pylint: disable=eval-used
            considering.extend([eval('fparser.two.Fortran2003.' + name)
                                for name in consideration.subclass_names])

    @staticmethod
    def print_tree(children, indent=0):
        '''
        Dumps a textual representation of the tree to standard out.
        Intended for debug use.
        '''
        for child in children:
            print(' ' * indent + child.__class__.__name__)
            if isinstance(child, fparser.two.Fortran2003.BlockBase):
                FortranSource.print_tree(child.content, indent+1)
            elif isinstance(child, fparser.two.Fortran2003.SequenceBase):
                FortranSource.print_tree(child.items, indent+1)


class CSource(SourceTree):
    '''
    Holds a C/C++ source file as both a text block and parse tree.
    '''
    def get_tree(self):
        raise NotImplementedError('C/C++ source is not supported yet.')

    def get_tree_error(self):
        raise NotImplementedError('C/C++ source is not supported yet.')


class _SourceChain(object):
    # pylint: disable=too-few-public-methods
    '''
    Holds the chain of objects needed to understand a particular file
    extension.
    '''
    def __init__(self, extension, parser, *preprocessors):
        if extension[0] == '.':
            extension = extension[1:]
        self.extension = extension

        if not issubclass(parser, SourceTree):
            message = 'Object "{0}" must inherit SourceTree to be a parser.'
            raise Exception(message.format(parser.__class__.__name__))
        self.parser = parser

        for candidate in preprocessors:
            if not issubclass(candidate, SourceText):
                message = 'Object "{0}" must inherit SourceText' \
                          + ' to be a preprocessor'
                raise Exception(message.format(candidate.__class__.__name__))
        self.preprocessors = preprocessors


class SourceFactory(object):
    '''
    Manaages the handling of source file. Knows what chains of objects are
    needed to each file extension.
    '''
    # It is hard to lay down hard and fast rules about what should go in the
    # _extension_map list. Languages standards do not generally specify an
    # extension since it is an OS level concept.
    #
    # The default list should be kept short and the two goals should be
    # "correctness" and "commodity".
    #
    # Correctness means not including a slew of extensions just because they
    # happen to be used. Stick to ones which are "correct" in that they are
    # well thought out.
    #
    # For instance it is not uncommon for people to use all sorts of extensions
    # for Fortran which encodes the version of the language they are using.
    # This is a poor idea as it leads to a proliferation of extension for no
    # obvious gain. What's more a file with a ".f03" extension may be coded
    # using only Fortran 95 features.
    #
    # Instead we follow the convention that ".f90" means "free format" source
    # while ".f" continues to mean "fixed format".
    #
    # On the other hand commonality means including the most commonly used
    # extensions. That is why both ".cc" and ".cpp" are listed for C++ source.
    #
    # If your particular application needs to support some odd-ball extensions
    # then it can use the add_extension() call.
    #
    _extension_map = {'f90': _SourceChain('f90', FortranSource),
                      'F90': _SourceChain('F90',
                                          FortranSource, FortranPreProcessor),
                      'f': _SourceChain('f', FortranSource),
                      'F': _SourceChain('F',
                                        FortranSource, FortranPreProcessor),
                      'c': _SourceChain('c', CSource, CPreProcessor),
                      'h': _SourceChain('h', CSource, CPreProcessor),
                      'cc': _SourceChain('cc', CSource, CPreProcessor),
                      'cpp': _SourceChain('cpp', CSource, CPreProcessor)}

    @classmethod
    def add_extension(cls, extension, source, *preprocessors):
        '''
        Adds a mapping between source file extension and source handling
        classes.

        extension: str() File extension.
        source: class(LanguageSource) Source handler class.
        *preprocessors: class(SourceProcessor) Preprocessors to apply.
        '''
        if extension in cls._extension_map:
            raise Exception('Extension "{}" already mapped to a handler'
                            .format(extension))

        cls._extension_map[extension] = _SourceChain(extension,
                                                     source,
                                                     *preprocessors)

    @classmethod
    def get_extensions(cls):
        '''
        Gets the list of file extensions recognised by the read_file() method.
        '''
        return cls._extension_map.keys()

    @classmethod
    def read_file(cls, source_file):
        '''
        Creates a Source object from a file.

        The file extension is used to determine the source type so this will
        not work on "files" which do not have filenames.
        '''
        if isinstance(source_file, str):
            filename = source_file
        else:
            filename = source_file.name

        ext = os.path.splitext(filename)[1][1:]
        if ext not in cls._extension_map:
            raise Exception('Source file extension "{0}" not in handler map'
                            .format(ext))

        chain = cls._extension_map[ext]
        reader = SourceFileReader(source_file)
        # Decorate reader
        for handler_class in chain.preprocessors:
            reader = handler_class(reader)

        return chain.parser(reader)



#!/usr/bin/env python
##############################################################################
# (c) Crown copyright 2019 Met Office. All rights reserved.
# The file LICENCE, distributed with this code, contains details of the terms
# under which the code may be used.
##############################################################################
'''
Core of the Fortran style checking tool.
'''
import logging

import stylist.source


class CheckEngine(object):
    # pylint: disable=too-few-public-methods
    '''
    Manages the checking of Fortran source files against style lists.
    '''

    def __init__(self, styles):
        '''
        Constructs a CheckEngine object from list of style lists.

        styles (list) Style object for each style list to use in checks.
        '''
        self._styles = styles

    def check(self, source_filename):
        '''
        Passes the eyes of all registered style lists over the source file.

        source_filename (string or file-like) Fortran source file to check.
        '''
        issues = []
        with open(source_filename, 'rt') as source_file:
            logging.getLogger(__name__).info('Examining: ' + source_filename)
            source = stylist.source.SourceFactory.read_file(source_file)
            for astyle in self._styles:
                for new_issue in astyle.check(source):
                    new_issue.set_filename(source_filename)
                    issues.append(new_issue)
        return issues



#!/usr/bin/env python
##############################################################################
# (c) Crown copyright 2019 Met Office. All rights reserved.
# The file LICENCE, distributed with this code, contains details of the terms
# under which the code may be used.
##############################################################################
'''
Extensible code style checker currently supporting Fortran, PSyclone DSL, etc
'''

__version__ = '0.1'



#!/usr/bin/env python
##############################################################################
# (c) Crown copyright 2019 Met Office. All rights reserved.
# The file LICENCE, distributed with this code, contains details of the terms
# under which the code may be used.
##############################################################################
'''
A collection of style rules.
'''

from abc import ABCMeta, abstractmethod
import logging
import re

import fparser.two.Fortran2003
import fparser.common.readfortran
from stylist.issue import Issue
from stylist.source import FortranSource


class Rule(object, metaclass=ABCMeta):
    # pylint: disable=too-few-public-methods
    '''
    Abstract parent of all rules.
    '''
    @abstractmethod
    def examine(self, subject):
        '''
        Examines the provided source code object for an issue.

        Returns a list of stylist.issue.Issue. Naturally this may be an
        empty list if all was well.
        '''
        message = 'Rule: {name}'.format(name=str(self.__class__.__name__))
        logging.getLogger(__name__).info(message)
        return []


class FortranCharacterset(Rule):
    # pylint: disable=too-few-public-methods
    '''
    Scans the source for characters which are not supported by Fortran.
    '''
    _FORTRAN_LETTER = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
    _FORTRAN_DIGIT = '0123456789'
    _FORTRAN_ALPHANUMERIC = _FORTRAN_LETTER + _FORTRAN_DIGIT + '_'
    _FORTRAN_SPECIALS = ' =+-*/\\()[]{},.:;!"%&~<>?\'`^|$#@'
    _FORTRAN_CHARACTERSET = _FORTRAN_ALPHANUMERIC + _FORTRAN_SPECIALS

    _APOSTROPHY = "'"
    _EXCLAMATION = '!'
    _NEWLINE = '\n'
    _QUOTE = '"'

    def examine(self, subject):
        # pylint: disable=too-many-branches
        '''
        Examines the source code for none Fortran characters.

        This is complicated by the fact that the source must consist of only
        certain characters except comments and strings. These may contain
        anything.
        '''
        issues = super(FortranCharacterset, self).examine(subject)

        text = subject.get_text()
        index = 0
        line = 1
        state = 'code'
        while index < len(text):
            character = text[index]
            if state == 'code':
                if character == self._NEWLINE:
                    line += 1
                elif character == self._EXCLAMATION:
                    state = 'comment'
                elif character == self._APOSTROPHY:
                    state = 'apostraphystring'
                elif character == self._QUOTE:
                    state = 'quotestring'
                elif character in self._FORTRAN_CHARACTERSET:
                    pass
                else:
                    description = "Found character {char} " \
                                  + "not in Fortran character set"
                    description = description.format(char=repr(character))
                    issues.append(Issue(description, line=line))
            elif state == 'comment':
                if character == self._NEWLINE:
                    line += 1
                    state = 'code'
            elif state == 'apostraphystring':
                if character == self._APOSTROPHY:
                    state = 'code'
            elif state == 'quotestring':
                if character == self._QUOTE:
                    state = 'code'
            else:
                raise Exception('Parser in unknown state: ' + state)
            index += 1

        return issues


class TrailingWhitespace(Rule):
    '''
    Scans the source for tailing whitespace.
    '''
    _TRAILING_SPACE_PATTERN = re.compile(r'\s+$')

    def examine(self, subject):
        '''
        Examines the text for white space at the end of lines.

        This includes empty lines.
        :param subject: File contents as Source object.
        :return: List of issues or empty list.
        '''
        issues = super(TrailingWhitespace, self).examine(subject)

        text = subject.get_text()
        line_tally = 0
        for line in text.splitlines():
            line_tally += 1
            match = self._TRAILING_SPACE_PATTERN.search(line)
            if match:
                description = 'Found trailing white space'
                issues.append(Issue(description, line=line_tally))

        return issues


class FortranRule(Rule, metaclass=ABCMeta):
    '''
    Parent for style rules pertaining to Fortran source.
    '''
    # pylint: disable=too-few-public-methods, abstract-method
    def examine(self, subject):
        issues = super(FortranRule, self).examine(subject)

        if not isinstance(subject, FortranSource):
            description = '"{0}" passed to a Fortran rule'
            raise Exception(description.format(subject.get_language()))

        if not subject.get_tree():
            description = 'Unable to perform {} as source didn\'t parse: {}'
            issues.append(Issue(description.format(self.__class__.__name__,
                                                   subject.get_tree_error())))
            return issues

        issues.extend(self.examine_fortran(subject))
        return issues

    @abstractmethod
    def examine_fortran(self, subject):
        '''
        Examines the provided Fortran source code object for an issue.

        Returns a list of stylist.issue.Issue objects.
        '''
        raise NotImplementedError()


class MissingImplicit(FortranRule):
    # pylint: disable=too-few-public-methods
    '''
    Catches cases where code blocks which could have an "implicit" statement
    don't.
    '''
    def __init__(self, default):
        '''
        Constructor taking a default implication.

        Todo: This rule was designed to check merely for the presence of an
        "implicit" statement but to leave the choice of implication to the
        user. It might also be desirable to allow this rule to enforce a
        specific implication. In which case the "default" argument may need a
        different name. A switch would be necessary to flip between "any" and
        "specific" modes.

        Todo: At the moment the default is not used. In the future we may want
        to allow rules to enforce style rather than simply reporting it. In
        this case the default would be used where none is present.
        '''
        assert default.lower() in ('none', 'private', 'public')
        self._default = default.lower()

    _NATURE_MAP = {fparser.two.Fortran2003.Program_Stmt: 'Program',
                   fparser.two.Fortran2003.Module_Stmt: 'Module',
                   fparser.two.Fortran2003.Subroutine_Stmt: 'Subroutine',
                   fparser.two.Fortran2003.Function_Stmt: 'Function'}

    def examine_fortran(self, subject):
        issues = []

        scope_units = subject.path('Program_Unit')
        scope_units.extend(subject.path(['Main_Program',
                                         'Internal_Subprogram_Part',
                                         'Internal_Subprogram']))
        scope_units.extend(subject.path(['Module',
                                         'Module_Subprogram_Part',
                                         'Module_Subprogram']))
        for scope in scope_units:
            scope_statement = subject.get_first_statement(root=scope)

            implication = subject.path(['Specification_Part',
                                        'Implicit_Part',
                                        'Implicit_Stmt'],
                                       root=scope.content[1:])
            if not implication:
                nature = MissingImplicit._NATURE_MAP[scope_statement.__class__]
                name = scope_statement.items[1]
                description = "{thing} '{name}' is missing " \
                              + "an implicit statement"
                description = description.format(thing=nature, name=name)
                issues.append(Issue(description))
        return issues


class MissingOnly(FortranRule):
    '''
    Catches cases where a "use" statement is present but has no "only" claus.
    '''
    def __init__(self, ignore=[]):
        '''
        Constructs a "MissingOnly" rule object taking a list of exception
        modules which are not required to have an "only" clause.
        '''
        assert isinstance(ignore, list)
        self._ignore = ignore

    def examine_fortran(self, subject):
        issues = []

        for statement in subject.find_all(fparser.two.Fortran2003.Use_Stmt):
            module = statement.items[2]
            onlies = statement.items[4]
            if str(module).lower() not in self._ignore:
                if onlies is None:
                    description = 'Usage of "{module}" without "only" clause.'
                    issues.append(Issue(description.format(module=module)))

        return issues



#!/usr/bin/env python
##############################################################################
# (c) Crown copyright 2019 Met Office. All rights reserved.
# The file LICENCE, distributed with this code, contains details of the terms
# under which the code may be used.
##############################################################################
'''
Issues found in the source.
'''


class Issue(object):
    # pylint: disable=too-few-public-methods
    '''
    Holds details pertaining to an issue with the source.
    '''
    def __init__(self, description, line=None, filename=None):
        self._filename = filename
        self._line = line
        self._description = description

    def __str__(self):
        string = ''
        if self._filename:
            string += '{filename}: '
        if self._line:
            string += '{line}: '
        string += '{description}'
        return string.format(filename=self._filename,
                             line=self._line,
                             description=self._description)

    def set_filename(self, filename):
        '''
        Attaches a filename to this issue.
        '''
        self._filename = filename
